<?php
/**
 * Events Manager Stripe Gateway
 *
 * @package   EM_Gateway_Stripe
 * @author    Sam De Francesco <sam@defrances.co>
 * @license   GPL-2.0+
 * @link      http://defrances.co
 * @copyright 2013 Sam De Francesco
 */

/**
 * Stripe Gateway class
 *
 * @package EM_Gateway_Stripe
 * @author  Sam De Francesco <sam@defrances.co>
 */
class EM_Gateway_Stripe extends EM_Gateway {
    /**
     * Gateway reference, which is used in various places for referencing gateway info. Use lowercase characters/numbers and underscores.
     *
     * @since    1.0.0
     *
     * @var string
     */
    var $gateway = 'stripe';

    /**
     * This will be what admins see as the gatweway name (e.g. Offline, PayPal, Authorize.net ...)
     *
     * @since    1.0.0
     *
     * @var string
     */
    var $title = 'Stripe';

    /**
     * The default status value your gateway assigns this booking. Default is 4, i.e. awaiting online payment
     *
     * @since    1.0.0
     *
     * @var int
     */
    var $status = 4;

    /**
     * Set this to any true value and this will trigger the em_my_bookings_booked_message function to override the status name of this booking when in progress.
     *
     * @since    1.0.0
     *
     * @var string
     */
    var $status_txt = 'Processing (Stripe)';

    /**
     * If your gateway supports the ability to pay without requiring further fields (e.g. credit card info), then you can set this to true.
     * 
     * You will automatically have the ability to show buttons within your gateway. It's up to you to change what happens after by 
     * overriding functions from EM_Gateway such as modifying booking_add or booking_form_feedback.
     *
     * @since    1.0.0
     *
     *  
     * @var boolean
     */
    var $button_enabled = false;

    /**
     * If your gateway is compatible with our Multiple Bookings Mode, then you can set this to true, otherwise your gateway won't be available for booking in this mode.
     *
     * @since    1.0.0
     *
     *  
     * @var boolean
     */
    var $supports_multiple_bookings = true;    

    /**
     * Some external gateways (e.g. PayPal IPNs) return information back to your site about payments, which allow you to automatically track refunds made outside Events Manager.
     * If you enable this to true, be sure to add an overriding handle_payment_return function to deal with the information sent by your gateway.
     *
     * @since    1.0.0
     *
     * @var unknown_type
     */
    var $payment_return = true;

    /**
     * Timeout limit for automatic user registration
     *
     * @since   1.0.0
     *
     * @var     string
     */
    var $registered_timer = 0;

    /**
     * Plugin version, used for cache-busting of style and script file references.
     *
     * @since   1.0.0
     *
     * @var     string
     */
    protected $version = '1.0.5';

    /**
     * Unique identifier for your plugin.
     *
     * Use this value (not the variable name) as the text domain when internationalizing strings of text. It should
     * match the Text Domain file header in the main plugin file.
     *
     * @since    1.0.0
     *
     * @var      string
     */
    protected $plugin_slug = 'events-manager-stripe';

    /**
     * Slug of the plugin screen.
     *
     * @since    1.0.0
     *
     * @var      string
     */
    protected $plugin_screen_hook_suffix = null;

    /**
     * Initialize the plugin by setting localization, filters, and administration functions.
     *
     * @since     1.0.0
     */
    public function __construct() {
        parent::__construct();

        if ($this->is_active()) {
            // Load plugin text domain
            add_action( 'init', array( $this, 'load_plugin_textdomain' ) );

            //Force SSL for booking submissions, since we have card info
            if ( ( !empty($_SERVER['HTTPS']) && $_SERVER['HTTPS'] != 'off' )
                || (get_option('em_' . $this->gateway . '_mode') == 'live')) {
                /* modify booking script, force SSL for all */
                add_filter( 'em_wp_localize_script', array( $this, 'em_wp_localize_script' ), 10, 1 );
                add_filter( 'em_booking_form_action_url', array( $this, 'force_ssl' ), 10, 1 );
            }
        }
    }

   /**
     * This function intercepts the previous booking form url from the javascript localized array of EM variables and
     * forces it to be an HTTPS url. 
     *
     * @since    1.0.0
     *
     * @param array $localized_array
     * @return array
     */
    public function em_wp_localize_script($localized_array) {
        $localized_array['bookingajaxurl'] = $this->force_ssl($localized_array['bookingajaxurl']);
        if ( ! is_admin() )  $localized_array['ajaxurl']        = $this->force_ssl($localized_array['ajaxurl']);
        return $localized_array;
    }

    /**
     * Triggered by the em_booking_add_stripe action, modifies the booking status if the event isn't free and also adds a filter to modify user feedback returned.
     *
     * @since    1.0.0
     *
     * @param EM_Event $EM_Event
     * @param EM_Booking $EM_Booking
     * @param boolean $post_validation
     */
    public function booking_add($EM_Event,$EM_Booking, $post_validation = false){
        $this->registered_timer = current_time('timestamp', 1);
        parent::booking_add($EM_Event, $EM_Booking, $post_validation);

        if ( $post_validation && empty($EM_Booking->booking_id) ) {
            if ( get_option('dbem_multiple_bookings') && get_class($EM_Booking) == 'EM_Multiple_Booking' ) {
                add_filter( 'em_multiple_booking_save', array( &$this, 'em_booking_save' ), 2, 2 );               
            } else {
                add_filter( 'em_booking_save', array( &$this, 'em_booking_save' ), 2, 2 );
            }               
        }
    }

    /**
     * Added to filters once a booking is added. Once booking is saved, we capture payment,
     * and approve the booking (saving a second time). If payment isn't approved, just delete
     * the booking and return false for save. 
     * 
     * @since    1.0.0
     *
     * @param bool $result
     * @param EM_Booking $EM_Booking
     */
    public function em_booking_save($result, $EM_Booking) {
        global $wpdb, $wp_rewrite, $EM_Notices;

        //make sure booking save was successful before we try anything
        if ($result) {
            if ($EM_Booking->get_price() > 0) {
                //handle results
                $capture = $this->__charge( $EM_Booking );
                if ($capture) {
                    //Set booking status, but no emails sent
                    if (!get_option('em_' . $this->gateway . '_manual_approval', false) || !get_option('dbem_bookings_approval')) {
                        $EM_Booking->set_status(1, false); //Approve
                    } else {
                        $EM_Booking->set_status(0, false); //Set back to normal "pending"
                    }
                } else {
                    //not good.... error inserted into booking in capture function. Delete this booking from db
                    if (!is_user_logged_in() && get_option('dbem_bookings_anonymous') && !get_option('dbem_bookings_registration_disable') && !empty($EM_Booking->person_id)) {
                        //delete the user we just created, only if created after em_booking_add filter is called (which is when a new user for this booking would be created)
                        $EM_Person = $EM_Booking->get_person();
                        if( strtotime($EM_Person->data->user_registered) >= $this->registered_timer ){
                            if( is_multisite() ){
                                include_once(ABSPATH.'/wp-admin/includes/ms.php');
                                wpmu_delete_user($EM_Person->ID);
                            }else{
                                include_once(ABSPATH.'/wp-admin/includes/user.php');
                                wp_delete_user($EM_Person->ID);
                            }
                            //remove email confirmation
                            global $EM_Notices;
                            $EM_Notices->notices['confirms'] = array();
                        }
                    }
                    $EM_Booking->delete();
                    return false;
                }
            }
        }
        return $result;
    }

    /**
     * Intercepts return JSON and adjust feedback messages when booking with this gateway. This filter is added only when the em_booking_add function is triggered by the em_booking_add filter.
     *
     * @since    1.0.0
     *
     * @param array $return
     * @param EM_Booking $EM_Booking
     * @return array
     */
    public function booking_form_feedback( $return, $EM_Booking ){
        // Double check $EM_Booking is an EM_Booking object and that we have a booking awaiting payment.
        if (!empty($return['result'])) {
            if (!empty($EM_Booking->booking_meta['gateway']) && $EM_Booking->booking_meta['gateway'] == $this->gateway && $EM_Booking->get_price() > 0) {
                $return['message'] = get_option('em_stripe_booking_feedback');
            } else {
                //returning a free message
                $return['message'] = get_option('em_stripe_booking_feedback_free');
            }
        } elseif (!empty($EM_Booking->booking_meta['gateway']) && $EM_Booking->booking_meta['gateway'] == $this->gateway && $EM_Booking->get_price() > 0) {
            //void this last authroization
            $this->__void($EM_Booking);
        }
        return $return;
    }

    /**
     * Outputs extra custom content e.g. information about this gateway or extra form fields to be requested if
     * this gateway is selected (not applicable with Quick Pay Buttons).
     */
    public function booking_form() {
        echo get_option('em_' . $this->gateway . '_form');
        include ( plugin_dir_path( __FILE__ ) . 'views/booking_form.php' );
    }

    /**
     * Called by $this->settings(), override this to output your own gateway options on this gateway settings page  
     */
    public function mysettings() {
        global $EM_options;
        include ( plugin_dir_path( __FILE__ ) . 'views/settings.php' );
    }
    
    /**
     * Run by EM_Gateways::handle_gateways_panel_updates() if this gateway has been updated. You should capture the values
     * of your new fields above and save them as options here.
     * return boolean 
     */
    public function update() {
        parent::update();

        $gateway_options = array(
            $this->gateway . "_mode" => $_REQUEST['mode'],
            $this->gateway . "_api_key" => $_REQUEST['api_key'],
            $this->gateway . "_api_test_key" => $_REQUEST['api_test_key'],
            $this->gateway . "_user_login" => $_REQUEST['user_login'],
            $this->gateway . "_email_customer" => ($_REQUEST['email_customer']),
            $this->gateway . "_header_email_receipt" => $_REQUEST['header_email_receipt'],
            $this->gateway . "_footer_email_receipt" => $_REQUEST['footer_email_receipt'],
            $this->gateway . "_md5_hash" => $_REQUEST['md5_hash'],
            $this->gateway . "_manual_approval" => $_REQUEST['manual_approval'],
            $this->gateway . "_booking_feedback" => wp_kses_data($_REQUEST['booking_feedback']),
            $this->gateway . "_booking_feedback_free" => wp_kses_data($_REQUEST['booking_feedback_free'])
        );

        foreach ($gateway_options as $key => $option) {
            update_option('em_' . $key, stripslashes($option));
        }

        //default action is to return true
        return true;
    }

    /**
     * Add extra placeholders to the booking email. Called by em_booking_output_placeholder filter, added in this object __construct() function.
     * 
     * @param string $result
     * @param EM_Booking $EM_Booking
     * @param string $placeholder
     * @param string $target
     * @return string
     */
    public function em_booking_output_placeholder( $result, $EM_Booking, $placeholder, $target='html' ) {    
        /* hook to add extra placeholders if needed */
        do_action('em_stripe_booking_output_placeholder', $result, $EM_Booking, $placeholder, $target); 

        return parent::em_booking_output_placeholder($result, $EM_Booking, $placeholder, $target);
    }
    
    /**
     * Handle stripe Webhook callbacks
     * 
     */
    public function handle_payment_return() {
        global $wpdb;


        // set api key
        $this->__set_api();

        // retrieve the request's body and parse it as JSON
        $body = @file_get_contents('php://input');
        $r = json_decode($body);
        $txn_id = $r->data->object->id;

        if (!empty($r->type)) {
            EM_Pro::log("Received webhook type {$r->type} from Stripe");

            if ($r->type == 'charge.refunded') {
                /*
                 * Note that the refund amount given is inclusive of past refunds, not just current one. Update the existing transaction instead.
                 * Credit webhooks have unique transaction ids however we can find the booking by invoice number / booking id then cancel the booking
                 * and record it as a new transaction with the new amount.
                 */
                $txn = $wpdb->get_row($wpdb->prepare("SELECT  transaction_id,
                                                              transaction_gateway_id,
                                                              transaction_total_amount,
                                                              booking_id
                                                        FROM  " . EM_TRANSACTIONS_TABLE . "
                                                       WHERE  transaction_gateway_id = %s
                                                         AND  transaction_gateway = %s
                                                    ORDER BY  transaction_total_amount DESC
                                                       LIMIT  1"  , $txn_id, $this->gateway), ARRAY_A);

                $EM_Booking = new EM_Booking( $txn['booking_id'] );
                if ( ! empty( $EM_Booking->booking_id ) ) {
                    $amount = ($r->data->object->amount - $amount - $r->data->object->amount_refunded) / 100; // convert to dollars 
                    $this->__update_transaction($EM_Booking, $amount, get_option('dbem_bookings_currency'), current_time('mysql'),
                                              $txn_id, __('Adjusted', 'em-pro'), 'Refunded $'. ($r->data->object->amount_refunded/100) );
                    echo "Transaction Processed";
                } else {
                    echo "Transaction not found"; //meaningful output
                }

            } elseif ($r->type == 'customer.created') {
                /* it would be nice to associate the customer token with a WP account in here, but account might not be created yet */
            } elseif ($r->type == 'charge.succeeded') {
                try {
                    /* figure out the WP user from supplied customer token */
                    $customer = $r->data->object->customer;
                    $cust_data = Stripe_Customer::retrieve($customer);
                    $email = $cust_data['email'];

                    /* From Email address, see if WP Account exists */
                    $user = get_user_by('email', $email);
                    $user_id = $user->ID;

                    /* If user exists, then store their customer token in their meta data */
                    if (isset($user_id)) {
                        $cc_token = get_user_meta($user_id, 'stripe_customer_token', true);
                        if (empty($cc_token)) {
                            add_user_meta($user_id, 'stripe_customer_token', $customer);
                        }
                    }

                    /* update or add transaction */
                    $txn_id = $r->data->object->id;
                    $txn = $wpdb->get_row( $wpdb->prepare( "
                                                        SELECT  transaction_id,
                                                                transaction_gateway_id,
                                                                transaction_total_amount,
                                                                booking_id
                                                          FROM  " . EM_TRANSACTIONS_TABLE . "
                                                         WHERE  transaction_gateway_id = %s
                                                           AND  transaction_gateway = %s
                                                      ORDER BY  transaction_total_amount DESC
                                                         LIMIT  1",
                                          $txn_id, $this->gateway ), ARRAY_A );

                    if (is_array($txn) && $txn['transaction_gateway_id'] == $txn_id && !empty($txn['booking_id'])) {
                        /* we know the booking id associated with this transaction, so update it */
                        $wpdb->update( EM_TRANSACTIONS_TABLE, array('transaction_status' => __('Confirmed', 'em-pro'),
                                      'transaction_timestamp' => current_time('mysql')), array('transaction_id' => $txn['transaction_id']));
                    } else {
                        /*
                         * Anonymous transaction, we don't know which event it applies to as there is no booking id. It was most likely a paymen initiated
                         * from the Stripe back end. If the user is known then we just associate the transaction with the last event they signed
                         * up for (XXX need to think about this assumption - what if they signed up for multiple events?). Otherwise, just ignore.
                         */
                        if (isset($user_id)) {
                            $booking_id = $wpdb->get_row($wpdb->prepare("SELECT  booking_id
                                                                           FROM  " . EM_BOOKINGS_TABLE . "
                                                                          WHERE  person_id = %s
                                                                       ORDER BY  booking_date DESC
                                                                          LIMIT  1"  , $user_id), ARRAY_A);
                            EM_Pro::log("Stripe anonymous webhook transaction $txn_id - user id is $user_id - assuming booking ID $booking_id");
                            $EM_Booking = new EM_Booking($booking_id);
                            $amount = $r->data->object->amount / 100;  // convert to dollar/pound amount
                            $this->record_transaction( $EM_Booking, $amount, get_option('dbem_bookings_currency'), current_time('mysql'),
                                                      $txn_id, __( 'Completed', 'em-pro' ), $r->data->object->description );
                        } else {
                            EM_Pro::log("Ignoring Stripe transaction $txn_id with no associated booking or user.");
                        }
                    }
                } catch (Exception $e) {
                    echo "Error retrieving user object: " . $e->getMessage();
                }
            } else {
                echo "Unhandled stripe webhook - " . $r->type;
            }
        } else {
            echo "Response not recognized.";
        } 
    }

    /**
     * Load the plugin text domain for translation.
     *
     * @since    1.0.0
     */
    public function load_plugin_textdomain() {

        $domain = $this->plugin_slug;
        $locale = apply_filters( 'plugin_locale', get_locale(), $domain );

        load_textdomain( $domain, WP_LANG_DIR . '/' . $domain . '/' . $domain . '-' . $locale . '.mo' );
        load_plugin_textdomain( $domain, FALSE, dirname( plugin_basename( __FILE__ ) ) . '/lang/' );
    }

    /**
     * Turns any url into an HTTPS url.
     *
     * @since    1.0.0
     *
     * @param string $url
     * @return string
     */
    public function force_ssl($url) {
        return str_replace('http://', 'https://', $url);
    }   

   /**
    * Retreive the stripe vars needed to send to the gateway to proceed with payment
    *
    * @since     1.0.0
    *
    * @param EM_Booking $EM_Booking
    */
    private function __charge($EM_Booking) {
        global $EM_Notices;
        global $current_user;

        $this->__set_api();

        /* check if account has customer token already, if not then create one */
        get_currentuserinfo();
        $user_id = $current_user->ID; //logged in user's ID

        $cc_token = get_user_meta($user_id, 'stripe_customer_token', true);

        try {
            $customer_params = array(
                "email" => $EM_Booking->get_person()->user_email,
                "card" => array(
                    "number" => $_REQUEST['x_card_num'],
                    "exp_month" => $_REQUEST['x_exp_date_month'],
                    "exp_year" => $_REQUEST['x_exp_date_year'],
                    "cvc" => $_REQUEST['x_card_code']
                ),
                "description" => preg_replace('/[^a-zA-Z0-9<@>_\-+.\s]/', "", $EM_Booking->get_person()->get_name() .
                                              ' <' . $EM_Booking->get_person()->user_email . '>')
            );

            if (empty($cc_token)) {
                /* create Stripe CC token for this user */  
                $cust_response = Stripe_Customer::create($customer_params);
                $cc_token = $cust_response->id;
            } else {
                try {
                    /* Retreive customer token to update CC info */
                    $cust_response = Stripe_Customer::retrieve($cc_token);
                } catch (Exception $f) {
                    /* Token doesn't exist (or error) so create new one */
                    $cust_response = Stripe_Customer::create($customer_params);
                    $cc_token = $cust_response->id;
                }
                /* update customer token with CC details */
                $cust_response->card = array(
                    "number" => $_REQUEST['x_card_num'],
                    "exp_month" => $_REQUEST['x_exp_date_month'],
                    "exp_year" => $_REQUEST['x_exp_date_year'],
                    "cvc" => $_REQUEST['x_card_code']
                );
                $cust_response->description = preg_replace('/[^a-zA-Z0-9<@>_\-+.\s]/', "", $EM_Booking->get_person()->get_name() .
                                                           ' <' . $EM_Booking->get_person()->user_email . '>');
                $cust_response->email = $EM_Booking->get_person()->user_email;
                $cust_response->save();
            }

            if (!empty($user_id)) {
                /* if user is logged in, we update their stripe token in meta data.
                   If not, we will catch it via the webhook callback */
                update_user_meta($user_id, 'stripe_customer_token', $cc_token);
            }

            $EM_Booking->booking_meta['stripe_customer_token'] = $cc_token;

            /* Get Payment */
            $amount = $EM_Booking->get_price(false, false, true);
            $charge_params = array(
                "amount" => $amount * 100, // convert to cents for Stripe
                "currency" => get_option('dbem_bookings_currency'),
                "customer" => $cc_token,
                "description" => preg_replace('/[^a-zA-Z0-9\s]/', "", $EM_Booking->get_event()->event_name)
            );
            $response = Stripe_Charge::create($charge_params);

            $EM_Booking->booking_meta[$this->gateway] = array('txn_id' => $response->id, 'amount' => $amount);
            $this->record_transaction($EM_Booking, $amount, get_option('dbem_bookings_currency'),
                                      date('Y-m-d H:i:s', current_time('timestamp')), $response->id, 'Completed', '');

            $result = true;
        } catch (Exception $e) {
            $EM_Booking->add_error($e->getMessage());
            EM_Pro::log("Stripe charge failed - " . $e->getMessage());
            $result = false;
        }

        //Return transaction_id or false
        return apply_filters('em_gateway_stripe_authorize', $result, $EM_Booking, $this);
    }

    /**
     * Get the Stripe object and set up basic parameters
     *
     * @since     1.0.0
     *
     */
    private function __set_api() {
        if (!class_exists('Stripe')) {
            require_once( plugin_dir_path( __FILE__ ) . 'lib/Stripe.php' );
        }

        if (get_option('em_' . $this->gateway . '_mode') == 'live') {
            $key = get_option('em_' . $this->gateway . '_api_key');
        } else {
            $key = get_option('em_' . $this->gateway . '_api_test_key');
        }

        //Basic Credentials
        Stripe::setApiKey($key);
    }   

    /**
     * Update or Insert transaction
     *
     * @since     1.0.0
     *
     * @param EM_Booking $EM_Booking
     * @param float $amount
     * @param string $currency
     * @param string $timestamp
     * @param string $txn_id
     * @param int $status
     * @param string $note
     */
    private function __update_transaction($EM_Booking, $amount, $currency, $timestamp, $txn_id, $status, $note) {
        global $wpdb;
        $data = array();
        $data['booking_id'] = $EM_Booking->booking_id;
        $data['transaction_gateway_id'] = $txn_id;
        $data['transaction_timestamp'] = $timestamp;
        $data['transaction_currency'] = $currency;
        $data['transaction_status'] = $status;
        $data['transaction_total_amount'] = $amount;
        $data['transaction_note'] = $note;
        $data['transaction_gateway'] = $this->gateway;

        if( !empty($txn_id) ) {
            $existing = $wpdb->get_row( $wpdb->prepare( "SELECT transaction_id,
                                                                transaction_status,
                                                                transaction_gateway_id,
                                                                transaction_total_amount
                                                           FROM " . EM_TRANSACTIONS_TABLE . "
                                                          WHERE transaction_gateway = %s
                                                            AND transaction_gateway_id = %s
                                                                ", $this->gateway, $txn_id ) );
        }

        $table = EM_TRANSACTIONS_TABLE;
        if ( is_multisite() && !EM_MS_GLOBAL && !empty( $EM_Event->blog_id ) && !is_main_site( $EM_Event->blog_id ) ) {
            //we must get the prefix of the transaction table for this event's blog if it is not the root blog
            $table = $wpdb->get_blog_prefix( $EM_Event->blog_id ) . 'em_transactions';
        }

        if( !empty($existing->transaction_gateway_id) ) {
            // Update existing
            $wpdb->update( $table, $data, array('transaction_id' => $existing->transaction_id) );
        } else {
            // Create new
            $wpdb->insert( $table, $data );
        }
    }

    /**
     * Void a booking
     *
     * @since    1.0.0
     *
     * @param EM_Booking $EM_Booking
     */
    private function __void($EM_Booking) {
        global $wpdb;

        if ($EM_Booking->booking_meta['gateway'] == $this->gateway) {
            $sql = $wpdb->prepare("DELETE
                                     FROM " . EM_BOOKINGS_TABLE . "
                                    WHERE booking_id=%d", $EM_Booking->booking_id);

            $result = $wpdb->query($sql);
            if ($result !== false) {
                //delete the tickets 
                $EM_Booking->get_tickets_bookings()->delete();
                $EM_Booking->previous_status = $EM_Booking->booking_status;
                $EM_Booking->booking_status = false;
                $EM_Booking->feedback_message = sprintf(__('%s deleted', 'dbem'), __('Booking', 'dbem'));
            } else
                $EM_Booking->add_error(sprintf(__('%s could not be deleted', 'dbem'), __('Booking', 'dbem')));
        }
    }
}

/* eof */
